from unittest.mock import Mock, patch

from okta_oauth2.backend import OktaBackend


def test_backend_authenticate_requires_code_and_nonce(rf):
    """
    the authenticate method on the custom backend requires both
    an auth code and a nonce. If either aren't provided then
    authenitcate should return None
    """
    backend = OktaBackend()
    assert backend.authenticate(rf) is None


def test_authenticate_returns_a_user(rf, django_user_model):
    """
    We can't do the real authentication but we do need to make sure a
    real user is returned from the backend authenticate method if the
    TokenValidator succeeds, so fake success and see what happens.
    """
    user = django_user_model.objects.create_user("testuser", "testuser@example.com")

    with patch(
        "okta_oauth2.backend.TokenValidator.tokens_from_auth_code",
        Mock(return_value=(user, None)),
    ):
        backend = OktaBackend()
        assert backend.authenticate(rf, auth_code="123456", nonce="imanonce") == user



from django.contrib.auth.backends import ModelBackend

from .conf import Config
from .tokens import TokenValidator

config = Config()


class OktaBackend(ModelBackend):
    """
    Uses the same user store as the django ModelBackend but actually
    does its authentication using Okta's OIDC authorization servers.

    The Okta sign in widget will accept a username and password,
    validate them, and if successful return an authorization code.

    We take that code and use it to obtain an Access Token, an
    ID Token and a Refresh Token from Okta, set them in the session,
    and get the user from the Django database.
    """

    def authenticate(self, request, auth_code=None, nonce=None):
        if auth_code is None or nonce is None:
            return

        validator = TokenValidator(config, nonce, request)
        user, tokens = validator.tokens_from_auth_code(auth_code)

        if self.user_can_authenticate(user):
            return user



from unittest.mock import Mock, patch

from django.http import HttpResponse
from django.urls import reverse
from okta_oauth2.exceptions import TokenExpired
from okta_oauth2.middleware import OktaMiddleware
from okta_oauth2.tests.utils import build_id_token, update_okta_settings


def test_no_token_redirects_to_login(rf):
    """
    If there's no token in the session then we should be
    redirecting to the login.
    """
    request = rf.get("/")
    request.session = {}
    mw = OktaMiddleware(Mock(return_value=HttpResponse()))
    response = mw(request)
    assert response.status_code == 302
    assert response.url == reverse("okta_oauth2:login")


def test_invalid_token_redirects_to_login(rf):
    """
    It there's a token but it's invalid we should be
    redirecting to the login.
    """
    request = rf.get("/")
    request.COOKIES["okta-oauth-nonce"] = "123456"
    request.session = {"tokens": {}}
    mw = OktaMiddleware(Mock(return_value=HttpResponse()))
    response = mw(request)
    assert response.status_code == 302
    assert response.url == reverse("okta_oauth2:login")


def test_valid_token_returns_response(rf):
    """
    If we have a valid token we should be returning the normal
    response from the middleware.
    """

    nonce = "123456"
    # We're building a token here that we know will be valid
    token = build_id_token(nonce=nonce)

    with patch(
        "okta_oauth2.middleware.TokenValidator._jwks", Mock(return_value="secret")
    ):
        request = rf.get("/")
        request.COOKIES["okta-oauth-nonce"] = nonce
        request.session = {"tokens": {"id_token": token}}
        mw = OktaMiddleware(Mock(return_value=HttpResponse()))
        response = mw(request)
        assert response.status_code == 200


def test_token_expired_triggers_refresh(rf):
    """
    Test that an expired token triggers
    an attempt at refreshing the token.
    """
    raises_token_expired = Mock()
    raises_token_expired.side_effect = TokenExpired

    with patch(
        "okta_oauth2.middleware.TokenValidator.validate_token", raises_token_expired
    ), patch("okta_oauth2.middleware.TokenValidator.tokens_from_refresh_token"):

        request = rf.get("/")
        request.COOKIES["okta-oauth-nonce"] = "123456"
        request.session = {
            "tokens": {
                "id_token": "imanexpiredtoken",
                "refresh_token": "imsorefreshing",
            }
        }
        mw = OktaMiddleware(Mock(return_value=HttpResponse()))
        response = mw(request)
        assert response.status_code == 200


def test_token_expired_triggers_refresh_with_no_refresh(rf):
    """
    Test that an expired token triggers
    an attempt at refreshing the token. In this situation we
    don't have a refresh token so we should be redirecting back
    to login.
    """
    raises_token_expired = Mock()
    raises_token_expired.side_effect = TokenExpired

    with patch(
        "okta_oauth2.middleware.TokenValidator.validate_token", raises_token_expired
    ), patch("okta_oauth2.middleware.TokenValidator.tokens_from_refresh_token"):

        request = rf.get("/")
        request.COOKIES["okta-oauth-nonce"] = "123456"
        request.session = {"tokens": {"id_token": "imanexpiredtoken"}}
        mw = OktaMiddleware(Mock(return_value=HttpResponse()))
        response = mw(request)
        assert response.status_code == 302
        assert response.url == reverse("okta_oauth2:login")


def test_middleware_allows_public_url(settings, rf):
    """
    A URL that has been defined as a public url
    should just pass through our middleware.
    """
    settings.OKTA_AUTH = update_okta_settings(
        settings.OKTA_AUTH, "PUBLIC_NAMED_URLS", ("named-url",)
    )
    request = rf.get("/named/")
    request.session = {}
    mw = OktaMiddleware(Mock(return_value=HttpResponse()))
    response = mw(request)
    assert response.status_code == 200


def test_unauthorized_post_returns_401(settings, rf):
    """
    redirecting a POST is bad form so just return
    a 401 Unauthorized response if no token is there.
    """
    request = rf.post("/named/")
    request.session = {}
    mw = OktaMiddleware(Mock(return_value=HttpResponse()))
    response = mw(request)
    assert response.status_code == 401



import re

import pytest
from django.core.exceptions import ImproperlyConfigured
from okta_oauth2.conf import Config
from okta_oauth2.tests.utils import update_okta_settings


def test_conf_raises_error_if_no_settings(settings):
    """
    if there's no OKTA_AUTH in settings then we should
    be raising an ImproperlyConfigured exception.
    """
    del settings.OKTA_AUTH
    with pytest.raises(ImproperlyConfigured):
        Config()


def test_public_named_urls_are_built(settings):
    """
    We should have reversed url regexes to match against
    in our config objects.
    """
    settings.OKTA_AUTH = update_okta_settings(
        settings.OKTA_AUTH, "PUBLIC_NAMED_URLS", ("named-url",)
    )
    config = Config()
    assert config.public_urls == [
        re.compile("^/named/$"),
        re.compile("^/accounts/login/$"),
        re.compile("^/accounts/logout/$"),
        re.compile("^/accounts/oauth2/callback/$"),
    ]


def test_invalid_public_named_urls_are_ignored(settings):
    """
    We don't want to crash if our public named urls don't
    exist, instead just skip it.
    """
    settings.OKTA_AUTH = update_okta_settings(
        settings.OKTA_AUTH, "PUBLIC_NAMED_URLS", ("not-a-valid-url",)
    )
    config = Config()
    assert config.public_urls == [
        re.compile("^/accounts/login/$"),
        re.compile("^/accounts/logout/$"),
        re.compile("^/accounts/oauth2/callback/$"),
    ]



import os

SECRET_KEY = "imasecretlol"

DATABASES = {"default": {"NAME": "test.db", "ENGINE": "django.db.backends.sqlite3"}}

INSTALLED_APPS = (
    "django.contrib.admin",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.sites",
    "okta_oauth2.apps.OktaOauth2Config",
)

OKTA_AUTH = {
    "ORG_URL": "https://test.okta.notreal/",
    "ISSUER": "https://test.okta.notreal/oauth2/default",
    "CLIENT_ID": "not-a-real-id",
    "CLIENT_SECRET": "not-a-real-secret",
    "REDIRECT_URI": "http://localhost:8000/accounts/callback/",
}

ROOT_URLCONF = "okta_oauth2.tests.urls"

AUTHENTICATION_BACKENDS = ("okta_oauth2.backend.OktaBackend",)

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "APP_DIRS": True,
        "DIRS": [os.path.join(os.path.dirname(__file__), "templates"),],
        "OPTIONS": {
            "context_processors": [
                # Django builtin
                "django.template.context_processors.debug",
                "django.template.context_processors.media",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
            ]
        },
    },
]



import base64
import time

import jwt as jwt_python
import requests
from django.contrib.auth import get_user_model
from django.core.cache import caches
from jose import jws, jwt
from jose.exceptions import JWSError, JWTError

from .exceptions import (
    InvalidClientID,
    InvalidTokenSignature,
    IssuerDoesNotMatch,
    NonceDoesNotMatch,
    TokenExpired,
    TokenRequestFailed,
    TokenTooFarAway,
)

UserModel = get_user_model()


class DiscoveryDocument:
    # Find the OIDC metadata through discovery
    def __init__(self, issuer_uri):
        r = requests.get(issuer_uri + "/.well-known/openid-configuration")
        self.json = r.json()

    def getJson(self):
        return self.json


class TokenValidator:
    _discovery_document = None

    def __init__(self, config, nonce, request):
        self.config = config
        self.cache = caches[config.cache_alias]
        self.cache_key = "{}-keys".format(config.cache_prefix)
        self.request = request
        self.nonce = nonce

    @property
    def discovery_document(self):
        if self._discovery_document is None:
            self._discovery_document = DiscoveryDocument(self.config.issuer)
        return self._discovery_document

    def tokens_from_auth_code(self, code):
        data = {"grant_type": "authorization_code", "code": str(code)}

        result = self.call_token_endpoint(data)
        return self.handle_token_result(result)

    def tokens_from_refresh_token(self, refresh_token):
        data = {"grant_type": "refresh_token", "refresh_token": str(refresh_token)}

        result = self.call_token_endpoint(data)
        return self.handle_token_result(result)

    def handle_token_result(self, token_result):
        tokens = {}

        if token_result is None or "id_token" not in token_result:
            return None, tokens

        claims = self.validate_token(token_result["id_token"])

        if claims:
            tokens["id_token"] = token_result["id_token"]
            tokens["claims"] = claims

            try:
                user = UserModel._default_manager.get_by_natural_key(claims["email"])
            except UserModel.DoesNotExist:
                if (
                    self.config.superuser_group
                    and "groups" in claims
                    and self.config.superuser_group in claims["groups"]
                ):
                    user = UserModel._default_manager.create_user(
                        username=claims["email"],
                        email=claims["email"],
                        is_staff=True,
                        is_superuser=True,
                    )
                else:
                    user = UserModel._default_manager.create_user(
                        username=claims["email"], email=claims["email"]
                    )

        if "access_token" in token_result:
            tokens["access_token"] = token_result["access_token"]

        if "refresh_token" in token_result:
            tokens["refresh_token"] = token_result["refresh_token"]

        if user:
            self.request.session["tokens"] = tokens
            self.request.session.modified = True

        return user, tokens

    def call_token_endpoint(self, endpoint_data):
        """ Call /token endpoint
            Returns access_token, id_token, and/or refresh_token
        """
        discovery_doc = self.discovery_document.getJson()
        token_endpoint = discovery_doc["token_endpoint"]

        basic_auth_str = "{0}:{1}".format(
            self.config.client_id, self.config.client_secret
        )
        authorization_header = base64.b64encode(basic_auth_str.encode())
        header = {
            "Authorization": "Basic: " + authorization_header.decode("utf-8"),
            "Content-Type": "application/x-www-form-urlencoded",
        }

        data = {"scope": self.config.scopes, "redirect_uri": self.config.redirect_uri}

        data.update(endpoint_data)
        # Send token request
        r = requests.post(token_endpoint, headers=header, params=data)
        response = r.json()

        # Return object
        result = {}
        if "error" not in response:
            if "access_token" in response:
                result["access_token"] = response["access_token"]
            if "id_token" in response:
                result["id_token"] = response["id_token"]
            if "refresh_token" in response:
                result["refresh_token"] = response["refresh_token"]
        else:
            raise TokenRequestFailed(
                response["error"], response.get("error_description", None)
            )

        return result if len(result.keys()) > 0 else None

    def request_jwks(self):
        discovery_doc = self.discovery_document.getJson()
        r = requests.get(discovery_doc["jwks_uri"])
        return r.json()

    def _jwks(self, kid):
        """
            Internal:
                Fetch public key from jwks_uri and caches it until the key rotates
            :param kid: "key Id"
            :return: key from jwks_uri having the kid key
        """

        cached_keys = self.cache.get(self.cache_key) or []

        for key in cached_keys:
            if key["kid"] == kid:
                return key

        # lookup the key from jwks_uri if key is not in cache
        jwks = self.request_jwks()

        for key in jwks["keys"]:
            if kid == key["kid"]:
                cached_keys.append(key)
                self.cache.set(self.cache_key, cached_keys, self.config.cache_timeout)
                return key

        return None

    def validate_token(self, token):
        """
        Validate token
        (Taken from
        http://openid.net/specs/openid-connect-core-1_0.html#TokenResponseValidation)
        """

        """ Step 1
            If encrypted, decrypt it using the keys and algorithms specified
            in the meta_data

            If encryption was negotiated but not provided, REJECT

            Skipping Okta has not implemented encrypted JWT
        """

        decoded_token = jwt_python.decode(token, verify=False)

        dirty_alg = jwt.get_unverified_header(token)["alg"]
        dirty_kid = jwt.get_unverified_header(token)["kid"]

        key = self._jwks(dirty_kid)
        if key:
            # Validate the key using jose-jws
            try:
                jws.verify(token, key, algorithms=[dirty_alg])
            except (JWTError, JWSError) as err:
                raise InvalidTokenSignature("Invalid token signature") from err
        else:
            raise InvalidTokenSignature("Unable to fetch public signing key")

        """ Step 2
            Issuer Identifier for the OpenID Provider (which is typically
            obtained during Discovery) MUST exactly match the value of the
            iss (issuer) Claim.
            Redundant, since we will validate in Step 3, the "iss" claim matches
            host we requested the token from
        """

        if decoded_token["iss"] != self.config.issuer:
            """ Step 3
                Client MUST validate:
                    aud (audience) contains the same `client_id` registered
                    iss (issuer) identified as the aud (audience)
                    aud (audience) Claim MAY contain an array with more than one
                    element (Currently NOT IMPLEMENTED by Okta)
                The ID Token MUST be rejected if the ID Token does not list the
                Client as a valid audience, or if it contains additional audiences
                not trusted by the Client.
            """
            raise IssuerDoesNotMatch("Issuer does not match")

        if decoded_token["aud"] != self.config.client_id:
            raise InvalidClientID("Audience does not match client_id")

        """ Step 6 : TLS server validation not implemented by Okta
            If ID Token is received via direct communication between Client and
            Token Endpoint, TLS server validation may be used to validate the
            issuer in place of checking token
            signature. MUST validate according to JWS algorithm specialized in JWT
            alg Header. MUST use keys provided.
        """

        """ Step 7
            The alg value SHOULD default to RS256 or sent in
            id_token_signed_response_alg param during Registration

            We don't need to test this. Okta always signs in RS256
        """

        """ Step 8 : Not implemented due to Okta configuration

            If JWT alg Header uses MAC based algorithm (HS256, HS384, etc) the
            octets of UTF-8 of the client_secret corresponding to the client_id
            are contained in the aud (audience) are used to validate the signature.
            For MAC based, if aud is multi-valued or if azp value is different
            than aud value - behavior is unspecified.
        """

        if decoded_token["exp"] < int(time.time()):
            """ Step 9
                The current time MUST be before the time represented by exp
            """
            raise TokenExpired

        if decoded_token["iat"] < (int(time.time()) - 100000):
            """ Step 10 - Defined 'too far away time' : approx 24hrs
                The iat can be used to reject tokens that were issued too far away
                from current time, limiting the time that nonces need to be stored
                to prevent attacks.
            """
            raise TokenTooFarAway("iat too far in the past ( > 1 day)")

        if self.nonce is not None and "nonce" in decoded_token:
            """ Step 11
                If a nonce value is sent in the Authentication Request,
                a nonce MUST be present and be the same value as the one
                sent in the Authentication Request. Client SHOULD check for
                nonce value to prevent replay attacks.
            """
            if self.nonce != decoded_token["nonce"]:
                raise NonceDoesNotMatch(
                    "nonce value does not match Authentication Request nonce"
                )

        """ Step 12:  Not implemented by Okta
            If acr was requested, check that the asserted Claim Value is appropriate
        """

        """ Step 13
            If auth_time was requested, check claim value and request
            re-authentication if too much time elapsed

            We relax this requirement during jwt validation. The Okta Session
            should be handled inside Okta

            See https://developer.okta.com/docs/api/resources/sessions
        """

        return decoded_token



from http.cookies import SimpleCookie
from unittest.mock import Mock, patch

from django.test import Client
from django.urls import reverse


def test_callback_without_messages():
    """
    The okta callback function should
    return a 500 with the error message from okta if
    the messages framework is not enabled.
    """
    c = Client()

    response = c.get(
        reverse("okta_oauth2:callback"),
        {
            "error": "invalid_scope",
            "error_description": "One or more scopes are not "
            "configured for the authorization server resource.",
        },
    )

    assert response.status_code == 500
    assert (
        response.content == b"One or more scopes are not configured for "
        b"the authorization server resource."
    )


def test_callback_redirects_on_error(settings):
    """
    The okta callback function should set a message
    and return a redirect if it gets an error.
    """

    # We need to set up django messages to actually set a message.
    settings.INSTALLED_APPS = settings.INSTALLED_APPS + ("django.contrib.messages",)
    settings.MIDDLEWARE = (
        "django.contrib.sessions.middleware.SessionMiddleware",
        "django.contrib.messages.middleware.MessageMiddleware",
    )
    settings.TEMPLATES = [
        {
            "BACKEND": "django.template.backends.django.DjangoTemplates",
            "DIRS": [],
            "APP_DIRS": True,
            "OPTIONS": {
                "context_processors": [
                    # Django builtin
                    "django.template.context_processors.debug",
                    "django.template.context_processors.media",
                    "django.template.context_processors.request",
                    "django.contrib.auth.context_processors.auth",
                    "django.contrib.messages.context_processors.messages",
                ]
            },
        }
    ]

    c = Client()

    response = c.get(
        reverse("okta_oauth2:callback"),
        {
            "error": "invalid_scope",
            "error_description": "One or more scopes are not configured"
            " for the authorization server resource.",
        },
    )

    assert response.status_code == 302
    assert response.url == reverse("okta_oauth2:login")


def test_callback_success(settings, django_user_model):
    """
    the callback method should authenticate successfully with
    an auth_code and nonce. We have to fake this because we can't hit
    okta with a fake auth code.
    """

    settings.MIDDLEWARE = ("django.contrib.sessions.middleware.SessionMiddleware",)

    nonce = "123456"

    user = django_user_model.objects.create_user("testuser", "testuser@example.com")

    with patch(
        "okta_oauth2.backend.TokenValidator.tokens_from_auth_code",
        Mock(return_value=(user, None)),
    ):
        c = Client()

        c.cookies = SimpleCookie(
            {"okta-oauth-state": "cookie-state", "okta-oauth-nonce": nonce}
        )

        response = c.get(
            reverse("okta_oauth2:callback"), {"code": "123456", "state": "cookie-state"}
        )

        assert response.status_code == 302
        assert response.url == "/"


def test_login_view(client):
    response = client.get(reverse("okta_oauth2:login"))
    assert response.status_code == 200
    assert "config" in response.context


def test_login_view_deletes_cookies(client):
    client.cookies = SimpleCookie(
        {"okta-oauth-state": "cookie-state", "okta-oauth-nonce": "123456"}
    )

    response = client.get(reverse("okta_oauth2:login"))

    assert response.status_code == 200
    assert response.cookies["okta-oauth-state"].value == ""
    assert (
        response.cookies["okta-oauth-state"]["expires"]
        == "Thu, 01 Jan 1970 00:00:00 GMT"
    )
    assert response.cookies["okta-oauth-nonce"].value == ""
    assert (
        response.cookies["okta-oauth-nonce"]["expires"]
        == "Thu, 01 Jan 1970 00:00:00 GMT"
    )


def test_callback_rejects_post(client):
    response = client.post(reverse("okta_oauth2:callback"))
    assert response.status_code == 400


def test_invalid_states_is_a_bad_request(client):
    client.cookies = SimpleCookie(
        {"okta-oauth-state": "cookie-state", "okta-oauth-nonce": "nonce"}
    )

    response = client.get(
        reverse("okta_oauth2:callback"), {"code": "123456", "state": "wrong-state"}
    )

    assert response.status_code == 400


def test_failed_authentication_redirects_to_login(client, settings, django_user_model):
    settings.MIDDLEWARE = ("django.contrib.sessions.middleware.SessionMiddleware",)

    nonce = "123456"

    # Creating a user to make sure there's actually one that *could* be returned.
    django_user_model.objects.create_user("testuser", "testuser@example.com")

    with patch("okta_oauth2.views.authenticate", Mock(return_value=None)):
        c = Client()

        c.cookies = SimpleCookie(
            {"okta-oauth-state": "cookie-state", "okta-oauth-nonce": nonce}
        )

        response = c.get(
            reverse("okta_oauth2:callback"), {"code": "123456", "state": "cookie-state"}
        )

        assert response.status_code == 302
        assert response.url == reverse("okta_oauth2:login")


def test_logout_view_returns_200(client, settings):
    settings.MIDDLEWARE = ("django.contrib.sessions.middleware.SessionMiddleware",)

    response = client.get(reverse("okta_oauth2:logout"))
    assert response.status_code == 302
    assert response.url == reverse("okta_oauth2:login")
